@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/ftp.c 1.36 2000/01/12 17:43:50 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/ftp.c 1.48 2001/05/26 16:20:48 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5d.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.6d.
   Functions for getting URLs using FTP.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -41,6 +41,11 @@ static char *proxy=NULL;
 static int server_ctrl=-1,      /*+ for the control connection to the server. +*/
            server_data=-1;      /*+ for the data connection to the server. +*/
 
+#if USE_ZLIB
+/*+ A flag to indicate that a compressed page was requested. +*/
+static int request_compressed=0;
+#endif
+
 /*+ A buffer to contain the reply +*/
 static char *bufferhead=NULL,   /*+ head. +*/
             *buffer=NULL,       /*+ body. +*/
@@ -68,43 +73,35 @@ static char *htmlise_dir_entry(char *line);
 char *FTP_Open(URL *Url)
 {
  char *msg=NULL;
- char *colon;
+ char *hoststr,*portstr;
  char *server_host=NULL;
  int server_port=Protocols[Protocol_FTP].defport;
 
  /* Sort out the host. */
 
- proxy=WhichProxy(Url->proto,Url->host);
+ proxy=ConfigStringURL(Proxies,Url);
  if(IsLocalNetHost(Url->host))
     proxy=NULL;
 
  if(proxy)
-   {
-    server_host=(char*)malloc(strlen(proxy)+1);
-    strcpy(server_host,proxy);
-   }
+    server_host=proxy;
  else
-   {
-    server_host=(char*)malloc(strlen(Url->host)+1);
-    strcpy(server_host,Url->host);
-   }
+    server_host=Url->host;
 
- if((colon=strchr(server_host,':')))
-   {
-    *colon++=0;
-    if(*colon)
-       server_port=atoi(colon);
-   }
+ SplitHostPort(server_host,&hoststr,&portstr);
+
+ if(portstr)
+    server_port=atoi(portstr);
 
  /* Open the connection. */
 
- server_ctrl=OpenClientSocket(server_host,server_port,ConnectTimeout);
+ server_ctrl=OpenClientSocket(hoststr,server_port);
  init_buffer(server_ctrl);
 
  if(server_ctrl==-1)
-    msg=PrintMessage(Warning,"Cannot open the FTP control connection to %s port %d; [%!s].",server_host,server_port);
+    msg=PrintMessage(Warning,"Cannot open the FTP control connection to %s port %d; [%!s].",hoststr,server_port);
 
- free(server_host);
+ RejoinHostPort(server_host,hoststr,portstr);
 
  return(msg);
 }
@@ -129,7 +126,7 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
  char *host,*mimetype=NULL;
  char *msg_reply=NULL;
  char *timestamp,sizebuf[32];
- int i,l,port_l,port_h;
+ int i,l,port;
  struct tm modtime;
  time_t mtime;
  char *user,*pass;
@@ -146,16 +143,30 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    {
     char *head;
 
-    MakeRequestAuthorised(proxy,request_head);
+    /* Make the request OK for a proxy. */
+
+    MakeRequestProxyAuthorised(proxy,request_head);
+
+    /* Add the compression header */
+
+#if USE_ZLIB
+    if(ConfigBooleanURL(RequestCompressedData,Url) && !NotCompressed(NULL,Url->path))
+      {
+       request_compressed=1;
+       AddToHeader(request_head,"Accept-Encoding","deflate; q=1.0, x-gzip; q=0.9, gzip; q=0.9, identity; q=0.1");
+      }
+#endif
+
+    /* Send the request. */
 
     head=HeaderString(request_head);
 
     PrintMessage(ExtraDebug,"Outgoing Request Head (to proxy)\n%s",head);
 
     if(write_string(server_ctrl,head)==-1)
-       msg=PrintMessage(Warning,"Failed to write to remote FTP proxy [%!s].");
+       msg=PrintMessage(Warning,"Failed to write to remote FTP proxy; [%!s].");
     if(request_body && write_data(server_ctrl,request_body->content,request_body->length)==-1)
-       msg=PrintMessage(Warning,"Failed to write to remote FTP proxy [%!s].");
+       msg=PrintMessage(Warning,"Failed to write to remote FTP proxy; [%!s].");
 
     free(head);
 
@@ -164,7 +175,7 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  /* Else Sort out the path. */
 
- path=URLDecode(Url->path,0);
+ path=URLDecodeGeneric(Url->path);
  if(path[strlen(path)-1]=='/')
    {
     path[strlen(path)-1]=0;
@@ -185,7 +196,15 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
     pass=Url->pass;
    }
  else
-    WhatFTPUserPass(Url->host,&user,&pass);
+   {
+    user=ConfigStringURL(FTPAuthUser,Url);
+    if(!user)
+       user=ConfigString(FTPUserName);
+
+    pass=ConfigStringURL(FTPAuthPass,Url);
+    if(!pass)
+       pass=ConfigString(FTPPassWord);
+   }
 
  /* send all the RFC959 commands. */
 
@@ -193,7 +212,7 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    str=read_line_or_timeout(server_ctrl,str);
     PrintMessage(ExtraDebug,"FTP: connected; got: %s",str);
 
     if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
@@ -229,13 +248,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"USER %s\r\n",user)==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'USER' command to remote FTP host [%!s].");
+    msg=PrintMessage(Warning,"Failed to write 'USER' command to remote FTP host; [%!s].");
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    str=read_line_or_timeout(server_ctrl,str);
     PrintMessage(ExtraDebug,"FTP: sent 'USER %s'; got: %s",user,str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -257,13 +276,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    {
     if(write_formatted(server_ctrl,"PASS %s\r\n",pass)==-1)
       {
-       msg=PrintMessage(Warning,"Failed to write 'PASS' command to remote FTP host [%!s].");
+       msg=PrintMessage(Warning,"Failed to write 'PASS' command to remote FTP host; [%!s].");
        return(msg);
       }
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+       str=read_line_or_timeout(server_ctrl,str);
        PrintMessage(ExtraDebug,"FTP: sent 'PASS %s'; got: %s",pass,str);
 
        if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
@@ -300,13 +319,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"CWD %s\r\n",*path?path:"/")==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host [%!s].");
+    msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    str=read_line_or_timeout(server_ctrl,str);
     PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",*path?path:"/",str);
 
     if(!file && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
@@ -344,13 +363,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    {
     if(write_formatted(server_ctrl,"CWD %s\r\n",file)==-1)
       {
-       msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host [%!s].");
+       msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
        return(msg);
       }
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+       str=read_line_or_timeout(server_ctrl,str);
        PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",file,str);
       }
     while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -393,13 +412,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"TYPE %c\r\n",file?'I':'A')==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'TYPE' command to remote FTP host [%!s].");
+    msg=PrintMessage(Warning,"Failed to write 'TYPE' command to remote FTP host; [%!s].");
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    str=read_line_or_timeout(server_ctrl,str);
     PrintMessage(ExtraDebug,"FTP: sent 'TYPE %c'; got: %s",file?'I':'A',str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -425,47 +444,47 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        if(write_formatted(server_ctrl,"SIZE %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'SIZE' command to remote FTP host [%!s].");
+          msg=PrintMessage(Warning,"Failed to write 'SIZE' command to remote FTP host; [%!s].");
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          str=read_line_or_timeout(server_ctrl,str);
           PrintMessage(ExtraDebug,"FTP: sent 'SIZE %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
-       if (atoi(str) == 213)
-          if (str[4])
-             sprintf(sizebuf, "%ld", (long)atoi(str+4));
+       if(atoi(str)==213)
+          if(str[4])
+             sprintf(sizebuf,"%ld",(long)atoi(str+4));
 
        if(write_formatted(server_ctrl,"MDTM %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'MDTM' command to remote FTP host [%!s].");
+          msg=PrintMessage(Warning,"Failed to write 'MDTM' command to remote FTP host; [%!s].");
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          str=read_line_or_timeout(server_ctrl,str);
           PrintMessage(ExtraDebug,"FTP: sent 'MDTM %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
-       if (atoi(str) == 213)
+       if(atoi(str)==213)
          {
-          int year, mon, mday, hour, min, sec;
+          int year,mon,mday,hour,min,sec;
 
-          if (sscanf(str, "%*d %4d%2d%2d%2d%2d%2d", &year, &mon, &mday, &hour, &min, &sec) == 6)
+          if(sscanf(str,"%*d %4d%2d%2d%2d%2d%2d",&year,&mon,&mday,&hour,&min,&sec)==6)
             {
-             memset(&modtime, 0, sizeof(modtime));
-             modtime.tm_year = year - 1900;
-             modtime.tm_mon = mon - 1;
-             modtime.tm_mday = mday;
-             modtime.tm_hour = hour;
-             modtime.tm_min = min;
-             modtime.tm_sec = sec;
+             memset(&modtime,0,sizeof(modtime));
+             modtime.tm_year=year-1900;
+             modtime.tm_mon=mon-1;
+             modtime.tm_mday=mday;
+             modtime.tm_hour=hour;
+             modtime.tm_min=min;
+             modtime.tm_sec=sec;
             }
          }
       }
@@ -473,52 +492,104 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  /* Create the data connection. */
 
- if(write_string(server_ctrl,"PASV\r\n")==-1)
+ if(write_string(server_ctrl,"EPSV\r\n")==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'PASV' command to remote FTP host [%!s].");
+    msg=PrintMessage(Warning,"Failed to write 'EPSV' command to remote FTP host; [%!s].");
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
-    PrintMessage(ExtraDebug,"FTP: sent 'PASV'; got: %s",str);
+    str=read_line_or_timeout(server_ctrl,str);
+    PrintMessage(ExtraDebug,"FTP: sent 'EPSV'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
- if(!str || atoi(str)!=227)
+ if(!str)
    {
-    if(str)
+    msg=PrintMessage(Warning,"No reply from FTP server to 'EPSV' command; timed out?");
+    return(msg);
+   }
+
+ if(atoi(str) == 229)
+   {
+    host=strchr(str,'(');
+    if(!host || sscanf(host+1,"%*c%*c%*c%d%*c",&port)!=1)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command",str);
+       msg=PrintMessage(Warning,"Got '%s' message after sending 'EPSV' command, cannot parse.",str);
+       return(msg);
+      }
+    if(SocketRemoteName(server_ctrl,&host,NULL,NULL))
+      {
+       msg=PrintMessage(Warning,"Cannot determine server address.");
+       return(msg);
       }
-    else
-       msg=PrintMessage(Warning,"No reply from FTP server to 'PASV' command; timed out?");
-    return(msg);
    }
-
- if((host=strchr(str,',')))
-   {
-    while(isdigit(*--host));
-    host++;
-   }
-
- if(!host || sscanf(host,"%*d,%*d,%*d,%*d%n,%d,%d",&l,&port_h,&port_l)!=2)
+ else if(atoi(str)!=500)
    {
     char *p=str+strlen(str)-1;
+
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command, cannot parse.",str);
+    msg=PrintMessage(Warning,"Got '%s' message after sending 'EPSV' command",str);
     return(msg);
    }
+ else
+   {
+    int port_l,port_h;
 
- host[l]=0;
- for(;l>0;l--)
-    if(host[l]==',')
-       host[l]='.';
+    /* Let's try PASV instead then */
 
- server_data=OpenClientSocket(host,port_l+256*port_h,ConnectTimeout);
+    if(write_string(server_ctrl,"PASV\r\n")==-1)
+      {
+       msg=PrintMessage(Warning,"Failed to write 'PASV' command to remote FTP host; [%!s].");
+       return(msg);
+      }
+
+    do
+      {
+       str=read_line_or_timeout(server_ctrl,str);
+       PrintMessage(ExtraDebug,"FTP: sent 'PASV'; got: %s",str);
+      }
+    while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+    if(!str || atoi(str)!=227)
+      {
+       if(str)
+         {
+          char *p=str+strlen(str)-1;
+          while(*p=='\n' || *p=='\r') *p--=0;
+          msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command",str);
+         }
+       else
+          msg=PrintMessage(Warning,"No reply from FTP server to 'PASV' command; timed out?");
+       return(msg);
+      }
+
+    if((host=strchr(str,',')))
+      {
+       while(isdigit(*--host));
+       host++;
+      }
+
+    if(!host || sscanf(host,"%*d,%*d,%*d,%*d%n,%d,%d",&l,&port_h,&port_l)!=2)
+      {
+       char *p=str+strlen(str)-1;
+       while(*p=='\n' || *p=='\r') *p--=0;
+       msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command, cannot parse.",str);
+       return(msg);
+      }
+
+    port=port_l+256*port_h;
+
+    host[l]=0;
+    for(;l>0;l--)
+       if(host[l]==',')
+          host[l]='.';
+   }
+
+ server_data=OpenClientSocket(host,port);
  init_buffer(server_data);
 
  if(server_data==-1)
@@ -531,17 +602,21 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(!strcmp(request_head->method,"GET"))
    {
+    char *command;
+
     if(file)
       {
+       command="RETR";
+
        if(write_formatted(server_ctrl,"RETR %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host [%!s].");
+          msg=PrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host; [%!s].");
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          str=read_line_or_timeout(server_ctrl,str);
           PrintMessage(ExtraDebug,"FTP: sent 'RETR %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -550,18 +625,37 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       }
     else
       {
-       if(write_string(server_ctrl,"LIST -a\r\n")==-1)
-         {
-          msg=PrintMessage(Warning,"Failed to write 'LIST -a' command to remote FTP host [%!s].");
-          return(msg);
-         }
+       int i;
 
-       do
+       for(i=0;i<2;i++)
          {
-          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
-          PrintMessage(ExtraDebug,"FTP: sent 'LIST -a'; got: %s",str);
+          if(i==0)
+             command="LIST -a";
+          else
+             command="LIST";
+
+          if(write_formatted(server_ctrl,"%s\r\n",command)==-1)
+            {
+             msg=PrintMessage(Warning,"Failed to write '%s' command to remote FTP host; [%!s].",command);
+             return(msg);
+            }
+
+          do
+            {
+             str=read_line_or_timeout(server_ctrl,str);
+             PrintMessage(ExtraDebug,"FTP: sent '%s'; got: %s",command,str);
+            }
+          while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
+
+          if(i==0 && str && (atoi(str)!=150 && atoi(str)!=125))
+            {
+             char *p=str+strlen(str)-1;
+             while(*p=='\n' || *p=='\r') *p--=0;
+             msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
+            }
+          else
+             break;
          }
-       while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
        mimetype="text/html";
       }
@@ -579,10 +673,10 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
          {
           char *p=str+strlen(str)-1;
           while(*p=='\n' || *p=='\r') *p--=0;
-          msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,file?"RETR":"LIST -a");
+          msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
          }
        else
-          msg=PrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",file?"RETR":"LIST -a");
+          msg=PrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",command);
        return(msg);
       }
    }
@@ -592,13 +686,13 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        if(write_formatted(server_ctrl,"STOR %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'STOR' command to remote FTP host [%!s].");
+          msg=PrintMessage(Warning,"Failed to write 'STOR' command to remote FTP host; [%!s].");
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+          str=read_line_or_timeout(server_ctrl,str);
           PrintMessage(ExtraDebug,"FTP: sent 'STOR %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -673,12 +767,12 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
        if(msg_reply)
          {
           char *old_msg_reply=msg_reply;
-          msg_reply=HTMLString(msg_reply);
+          msg_reply=HTMLString(msg_reply,0);
           free(old_msg_reply);
          }
 
        buffer=HTMLMessageBody(-1,"FTPDir-Head",
-                              "url",URLDecode(Url->name,0),
+                              "url",Url->name,
                               "message",msg_reply,
                               NULL);
 
@@ -689,7 +783,7 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
  else
    {
     buffer=HTMLMessageBody(-1,"FTPPut",
-                           "url",URLDecode(Url->name,0),
+                           "url",Url->name,
                            NULL);
 
    }
@@ -726,7 +820,37 @@ char *FTP_ReadHead(char *line)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
-    return(read_line_or_timeout(server_ctrl,line,SocketTimeout));
+   {
+    char *newline;
+
+#if USE_ZLIB
+   repeat:
+#endif
+
+    newline=read_line_or_timeout(server_ctrl,line);
+
+#if USE_ZLIB
+    if(newline && request_compressed && !strncmp(newline,"Content-Encoding:",17))
+      {
+       int compression=WhichCompression(newline+17);
+
+       if(compression)
+         {
+          char *p=newline+17,*q=newline+strlen(newline)-1;
+          while(isspace(*p))
+             p++;
+          while(isspace(*q))
+             *q--=0;
+          PrintMessage(Inform,"Server has used 'Content-Encoding: %s'.",p); /* Used in audit-usage.pl */
+          init_zlib_buffer(server_ctrl,compression);
+          line=newline;
+          goto repeat;
+         }
+      }
+#endif
+
+    return(newline);
+   }
  
  /* Else send the header. */
 
@@ -765,7 +889,7 @@ int FTP_ReadBody(char *s,int n)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
-    return(read_data_or_timeout(server_ctrl,s,n,SocketTimeout));
+    return(read_data_or_timeout(server_ctrl,s,n));
  
  /* Else send the data then the tail. */
 
@@ -778,7 +902,7 @@ int FTP_ReadBody(char *s,int n)
        s[m]=buffertail[nreadtail];
    }
  else if(!buffer && !buffertail) /* File not dir entry. */
-    m=read_data_or_timeout(server_data,s,n,SocketTimeout);
+    m=read_data_or_timeout(server_data,s,n);
  else if(buffer && buffertail)  /* Middle of dir entry */
    {
     for(;nread<nbuffer && m<n;nread++,m++)
@@ -824,7 +948,13 @@ int FTP_Close(void)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
+   {
+#if USE_ZLIB
+    finish_zlib_buffer(server_ctrl);
+#endif
+
     return(CloseSocket(server_ctrl));
+   }
 
  /* Else say goodbye and close all of the sockets, */
 
@@ -835,7 +965,7 @@ int FTP_Close(void)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str,SocketTimeout);
+    str=read_line_or_timeout(server_ctrl,str);
     PrintMessage(ExtraDebug,"FTP: sent 'QUIT'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
@@ -868,7 +998,7 @@ static char *htmlise_dir_entry(char *line)
  int i,isdir=0,islink=0;
  char *q,*p[16],*l,*file=NULL,*link=NULL;
 
- l=read_line_or_timeout(server_data,line,SocketTimeout);
+ l=read_line_or_timeout(server_data,line);
 
  if(!l)
     return(l);
@@ -918,17 +1048,18 @@ static char *htmlise_dir_entry(char *line)
     endlc=*endl;
 
     *endf=0;
-    fileurlenc=URLEncode(file);
+    fileurlenc=URLEncodePath(file);
     *endf=endfc;
 
     *endl=0;
-    linkurlenc=URLEncode(link);
+    linkurlenc=URLEncodePath(link);
     *endl=endlc;
 
     ll=(char*)malloc(strlen(l)+strlen(fileurlenc)+strlen(linkurlenc)+40);
 
     strncpy(ll,l,file-l);
     strcpy(ll+(file-l),"<a href=\"");
+    strcat(ll,"./");
     strcat(ll,fileurlenc);
     if(isdir && *(endl-1)!='/')
        strcat(ll,"/");
@@ -942,6 +1073,8 @@ static char *htmlise_dir_entry(char *line)
       {
        strncat(ll,endf,link-endf);
        strcat(ll,"<a href=\"");
+       if(strncmp(linkurlenc,"../",3) && strncmp(linkurlenc,"./",2) && strncmp(linkurlenc,"/",1))
+          strcat(ll,"./");
        strcat(ll,linkurlenc);
        strcat(ll,"\">");
        *endl=0;
