@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/spool.c 2.42 1999/03/20 17:34:26 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/spool.c 2.57 2001/03/31 19:39:47 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4c.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.6c.
   Handle all of the spooling of files in the spool directory.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -38,6 +38,10 @@
 #endif
 
 
+/*+ The file descriptor of the sppol directory. +*/
+extern int fSpoolDir;
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Open a file in the outgoing directory to write into / read from.
 
@@ -56,10 +60,10 @@ int OpenOutgoingSpoolFile(int rw)
 
  /* Create the outgoing directory if needed and change to it */
 
- if(lstat("outgoing",&buf))
+ if(stat("outgoing",&buf))
    {
     PrintMessage(Inform,"Directory 'outgoing' does not exist [%!s]; creating one.");
-    if(mkdir("outgoing",0755))
+    if(mkdir("outgoing",(mode_t)ConfigInteger(DirPerm)))
       {PrintMessage(Warning,"Cannot create directory 'outgoing' [%!s].");return(-1);}
    }
  else
@@ -77,11 +81,11 @@ int OpenOutgoingSpoolFile(int rw)
     DIR *dir=opendir(".");
 
     if(!dir)
-      {PrintMessage(Warning,"Cannot open current directory 'outgoing' [%!s].");chdir("..");return(-1);}
+      {PrintMessage(Warning,"Cannot open current directory 'outgoing' [%!s].");fchdir(fSpoolDir);return(-1);}
 
     ent=readdir(dir);  /* skip .  */
     if(!ent)
-      {PrintMessage(Warning,"Cannot read current directory 'outgoing' [%!s].");closedir(dir);chdir("..");return(-1);}
+      {PrintMessage(Warning,"Cannot read current directory 'outgoing' [%!s].");closedir(dir);fchdir(fSpoolDir);return(-1);}
     ent=readdir(dir);  /* skip .. */
 
     while((ent=readdir(dir)))
@@ -91,7 +95,10 @@ int OpenOutgoingSpoolFile(int rw)
              PrintMessage(Inform,"Cannot rename file 'outgoing/%s' to 'outgoing/%s' [%!s]; race condition?",ent->d_name,name);
           else
             {
-             if((fd=open(name,O_RDONLY|O_BINARY))==-1)
+             fd=open(name,O_RDONLY|O_BINARY);
+             init_buffer(fd);
+
+             if(fd==-1)
                 PrintMessage(Inform,"Cannot open file 'outgoing/%s' [%!s]; race condition?",name);
              else
                {
@@ -106,14 +113,16 @@ int OpenOutgoingSpoolFile(int rw)
    }
  else
    {
-    fd=open(name,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,0644);
+    fd=open(name,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+    init_buffer(fd);
+
     if(fd==-1)
        PrintMessage(Warning,"Cannot open file 'outgoing/%s' [%!s]",name);
    }
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  return(fd);
 }
@@ -147,7 +156,10 @@ void CloseOutgoingSpoolFile(int fd,URL *Url)
  *newname='U';
 
  unlink(newname);
- ufd=open(newname,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,0644);
+
+ ufd=open(newname,O_WRONLY|O_CREAT|O_EXCL|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+ init_buffer(ufd);
+
  if(ufd!=-1)
    {
     write_string(ufd,Url->file);
@@ -160,7 +172,7 @@ void CloseOutgoingSpoolFile(int fd,URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(newname);
 }
@@ -178,7 +190,7 @@ int ExistsOutgoingSpoolFile(URL *Url)
 {
  struct stat buf;
  char *name;
- int exists=0;
+ int existsO,existsU;
 
  if(chdir("outgoing"))
    {PrintMessage(Warning,"Cannot change to directory 'outgoing' [%!s].");return(0);}
@@ -186,17 +198,20 @@ int ExistsOutgoingSpoolFile(URL *Url)
  /* Stat the outgoing file */
 
  name=URLToFileName(Url);
- *name='O';
 
- exists=!stat(name,&buf);
+ *name='O';
+ existsO=!stat(name,&buf);
+
+ *name='U';
+ existsU=!stat(name,&buf);
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(name);
 
- return(exists);
+ return(existsO&&existsU);
 }
 
 
@@ -222,24 +237,26 @@ char *HashOutgoingSpoolFile(URL *Url)
  *name='O';
 
  fd=open(name,O_RDONLY|O_BINARY);
+ init_buffer(fd);
+
  if(fd==-1)
-   {PrintMessage(Warning,"Cannot open outgoing request to create hash [%!s].");free(name);chdir("..");return(NULL);}
+   {PrintMessage(Warning,"Cannot open outgoing request to create hash [%!s].");free(name);fchdir(fSpoolDir);return(NULL);}
 
- req=(char*)malloc(256+1);
+ req=(char*)malloc(1024+1);
 
- r=read(fd,req,256);
- if(r==256)
+ r=read_data(fd,req,1024);
+ if(r==1024)
    {
     int rr=0;
     do
       {
        r+=rr;
-       req=(char*)realloc(req,r+256+1);
+       req=(char*)realloc(req,r+1024+1);
       }
-    while((rr=read(fd,&req[r],256))>0);
+    while((rr=read_data(fd,&req[r],1024))>0);
    }
  else if(r==-1)
-   {PrintMessage(Warning,"Cannot read from outgoing request to create hash [%!s].");close(fd);free(name);free(req);chdir("..");return(NULL);}
+   {PrintMessage(Warning,"Cannot read from outgoing request to create hash [%!s].");close(fd);free(name);free(req);fchdir(fSpoolDir);return(NULL);}
 
  req[r]=0;
 
@@ -249,7 +266,7 @@ char *HashOutgoingSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(name);
  free(req);
@@ -298,11 +315,11 @@ char *DeleteOutgoingSpoolFile(URL *Url)
     DIR *dir=opendir(".");
 
     if(!dir)
-      {err=PrintMessage(Warning,"Cannot open current directory 'outgoing' [%!s].");chdir("..");return(err);}
+      {err=PrintMessage(Warning,"Cannot open current directory 'outgoing' [%!s].");fchdir(fSpoolDir);return(err);}
 
     ent=readdir(dir);  /* skip .  */
     if(!ent)
-      {err=PrintMessage(Warning,"Cannot read current directory 'outgoing' [%!s].");closedir(dir);chdir("..");return(err);}
+      {err=PrintMessage(Warning,"Cannot read current directory 'outgoing' [%!s].");closedir(dir);fchdir(fSpoolDir);return(err);}
     ent=readdir(dir);  /* skip .. */
 
     while((ent=readdir(dir)))
@@ -316,7 +333,7 @@ char *DeleteOutgoingSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  return(err);
 }
@@ -340,10 +357,10 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
 
  /* Create the spool directory if needed and change to it. */
 
- if(lstat(Url->proto,&buf))
+ if(stat(Url->proto,&buf))
    {
     PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",Url->proto);
-    if(mkdir(Url->proto,0755))
+    if(mkdir(Url->proto,(mode_t)ConfigInteger(DirPerm)))
       {PrintMessage(Warning,"Cannot create directory '%s' [%!s].",Url->proto);return(-1);}
    }
  else
@@ -353,18 +370,18 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return(-1);}
 
- if(lstat(Url->host,&buf))
+ if(stat(Url->dir,&buf))
    {
-    PrintMessage(Inform,"Directory '%s/%s' does not exist [%!s]; creating one.",Url->proto,Url->host);
-    if(mkdir(Url->host,0755))
-      {PrintMessage(Warning,"Cannot create directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(-1);}
+    PrintMessage(Inform,"Directory '%s/%s' does not exist [%!s]; creating one.",Url->proto,Url->dir);
+    if(mkdir(Url->dir,(mode_t)ConfigInteger(DirPerm)))
+      {PrintMessage(Warning,"Cannot create directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return(-1);}
    }
  else
     if(!S_ISDIR(buf.st_mode))
-      {PrintMessage(Warning,"The file '%s/%s' is not a directory.",Url->proto,Url->host);chdir("..");return(-1);}
+      {PrintMessage(Warning,"The file '%s/%s' is not a directory.",Url->proto,Url->dir);fchdir(fSpoolDir);return(-1);}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(-1);}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return(-1);}
 
  /* Open the file for the web page. */
 
@@ -374,14 +391,17 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
  if(rw)
     fd=open(file,O_RDONLY|O_BINARY);
  else
-    fd=open(file,O_RDWR|O_CREAT|O_BINARY,0644);
+    fd=open(file,O_RDWR|O_CREAT|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+
+ init_buffer(fd);
 
  if(!rw && fd!=-1)
    {
     int ufd;
 
     *file='U';
-    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0644);
+    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+    init_buffer(ufd);
 
     if(ufd!=-1)
       {
@@ -400,14 +420,15 @@ int OpenWebpageSpoolFile(int rw,URL *Url)
  if(!rw && fd!=-1)
     utime(".",NULL);
 
- chdir("..");
+ fchdir(fSpoolDir);
+ chdir(Url->proto);
 
  if(!rw && fd!=-1)
     utime(".",NULL);
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(file);
 
@@ -434,8 +455,8 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
  if(chdir(Url->proto))
    {err=PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return(err);}
 
- if(chdir(Url->host))
-   {err=PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return(err);}
+ if(chdir(Url->dir))
+   {err=PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return(err);}
 
  /* Delete the file for the web page. */
 
@@ -445,11 +466,11 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
     DIR *dir=opendir(".");
 
     if(!dir)
-      {err=PrintMessage(Warning,"Cannot open current directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("../..");return(err);}
+      {err=PrintMessage(Warning,"Cannot open current directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return(err);}
 
     ent=readdir(dir);  /* skip .  */
     if(!ent)
-      {err=PrintMessage(Warning,"Cannot read current directory '%s/%s' [%!s].",Url->proto,Url->host);closedir(dir);chdir("../..");return(err);}
+      {err=PrintMessage(Warning,"Cannot read current directory '%s/%s' [%!s].",Url->proto,Url->dir);closedir(dir);fchdir(fSpoolDir);return(err);}
     ent=readdir(dir);  /* skip .. */
 
     while((ent=readdir(dir)))
@@ -462,13 +483,12 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
             {
              URL *delUrl=SplitURL(delurl);
 
-             chdir("../..");
+             fchdir(fSpoolDir);
 
-             if(ExistsLastTimeSpoolFile(delUrl))
-                DeleteLastTimeSpoolFile(delUrl);
+             DeleteLastTimeSpoolFile(delUrl);
 
              chdir(Url->proto);
-             chdir(Url->host);
+             chdir(Url->dir);
 
              free(delurl);
              FreeURL(delUrl);
@@ -476,17 +496,18 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
          }
 
        if(unlink(ent->d_name))
-          err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,ent->d_name);
+          err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,ent->d_name);
       }
 
     closedir(dir);
 
-    chdir("..");
+    fchdir(fSpoolDir);
+    chdir(Url->proto);
 
-    if(rmdir(Url->host))
-      err=PrintMessage(Warning,"Cannot delete what should be an empty directory '%s/%s' [%!s].",Url->proto,Url->host);
+    if(rmdir(Url->dir))
+      err=PrintMessage(Warning,"Cannot delete what should be an empty directory '%s/%s' [%!s].",Url->proto,Url->dir);
 
-    chdir("..");
+    fchdir(fSpoolDir);
    }
  else
    {
@@ -495,13 +516,13 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
     int didstat=1;
 
     if(stat(".",&buf))
-       PrintMessage(Warning,"Cannot stat directory '%s/%s' [%!s].",Url->proto,Url->host);
+       PrintMessage(Warning,"Cannot stat directory '%s/%s' [%!s].",Url->proto,Url->dir);
     else
        didstat=1;
 
     *file='D';
     if(unlink(file))
-       err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,file);
+       err=PrintMessage(Warning,"Cannot unlink cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,file);
 
     *file='U';
     unlink(file);
@@ -517,10 +538,9 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
        utime(".",&utbuf);
       }
 
-    chdir("../..");
+    fchdir(fSpoolDir);
 
-    if(ExistsLastTimeSpoolFile(Url))
-       DeleteLastTimeSpoolFile(Url);
+    DeleteLastTimeSpoolFile(Url);
    }
 
  return(err);
@@ -531,9 +551,11 @@ char *DeleteWebpageSpoolFile(URL *Url,int all)
   Touch a file in a spool subdirectory.
 
   URL *Url The URL to touch.
+
+  time_t when The time to set the access time to.
   ++++++++++++++++++++++++++++++++++++++*/
 
-void TouchWebpageSpoolFile(URL *Url)
+void TouchWebpageSpoolFile(URL *Url,time_t when)
 {
  char *file;
 
@@ -542,19 +564,37 @@ void TouchWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Touch the file for the web page. */
 
  file=URLToFileName(Url);
 
  *file='D';
- utime(file,NULL);
+
+ if(when)
+   {
+    struct stat buf;
+
+    if(stat(file,&buf))
+       utime(file,NULL);
+    else
+      {
+       struct utimbuf ubuf;
+
+       ubuf.actime=when;
+       ubuf.modtime=buf.st_mtime;
+
+       utime(file,&ubuf);
+      }
+   }
+ else
+    utime(file,NULL);
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(file);
 }
@@ -563,39 +603,45 @@ void TouchWebpageSpoolFile(URL *Url)
 /*++++++++++++++++++++++++++++++++++++++
   Check if a file in a spool subdirectory exists.
 
-  int ExistsWebpageSpoolFile Return a true value if the page exists.
+  time_t ExistsWebpageSpoolFile Return a the time the page was last accessed if the page exists.
 
   URL *Url The URL to check for.
   ++++++++++++++++++++++++++++++++++++++*/
 
-int ExistsWebpageSpoolFile(URL *Url)
+time_t ExistsWebpageSpoolFile(URL *Url)
 {
  struct stat buf;
  char *file;
- int exists=0;
+ int existsD,existsU;
 
  /* Change to the spool directory. */
 
  if(chdir(Url->proto))
     return(0);
 
- if(chdir(Url->host))
-   {chdir("..");return(0);}
+ if(chdir(Url->dir))
+   {fchdir(fSpoolDir);return(0);}
 
  /* Stat the file for the web page. */
 
  file=URLToFileName(Url);
- *file='D';
 
- exists=!stat(file,&buf);
+ *file='U';
+ existsU=!stat(file,&buf);
+
+ *file='D';
+ existsD=!stat(file,&buf);
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(file);
 
- return(exists);
+ if(existsU&&existsD)
+    return(buf.st_atime);
+ else
+    return(0);
 }
 
 
@@ -615,8 +661,8 @@ void CreateBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Create the filenames and rename the files. */
 
@@ -633,7 +679,7 @@ void CreateBackupWebpageSpoolFile(URL *Url)
  else
    {
     if(rename(orgfile,bakfile))
-       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->host,bakfile,orgfile);
+       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->dir,bakfile,orgfile);
 
     *bakfile=*orgfile='U';
     rename(orgfile,bakfile);
@@ -641,7 +687,7 @@ void CreateBackupWebpageSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(bakfile);
  free(orgfile);
@@ -664,8 +710,8 @@ void RestoreBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Create the filenames and rename the files. */
 
@@ -680,7 +726,7 @@ void RestoreBackupWebpageSpoolFile(URL *Url)
  if(!stat(bakfile,&buf))
    {
     if(rename(bakfile,orgfile))
-       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->host,bakfile,orgfile);
+       PrintMessage(Warning,"Cannot rename backup cached file '%s/%s/%s' to %s [%!s].",Url->proto,Url->dir,bakfile,orgfile);
 
     *bakfile=*orgfile='U';
     rename(bakfile,orgfile);
@@ -688,7 +734,7 @@ void RestoreBackupWebpageSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(bakfile);
  free(orgfile);
@@ -710,8 +756,8 @@ void DeleteBackupWebpageSpoolFile(URL *Url)
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Delete the file for the backup web page. */
 
@@ -721,14 +767,14 @@ void DeleteBackupWebpageSpoolFile(URL *Url)
  strcat(bakfile,"~");
 
  if(unlink(bakfile))
-    PrintMessage(Warning,"Cannot unlink backup cached file '%s/%s/%s' [%!s].",Url->proto,Url->host,bakfile);
+    PrintMessage(Warning,"Cannot unlink backup cached file '%s/%s/%s' [%!s].",Url->proto,Url->dir,bakfile);
 
  *bakfile='U';
  unlink(bakfile);
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(bakfile);
 }
@@ -743,39 +789,44 @@ void DeleteBackupWebpageSpoolFile(URL *Url)
 void CreateLockWebpageSpoolFile(URL *Url)
 {
  struct stat buf;
- char *lokfile,*orgfile;
+ char *lockfile,*orgfile;
+
+ /* Check for configuration file option. */
+
+ if(!ConfigBoolean(LockFiles))
+    return;
 
  /* Change to the spool directory. */
 
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Create the lock file for the web page. */
 
  orgfile=URLToFileName(Url);
 
- lokfile=(char*)malloc(strlen(orgfile)+2);
- strcpy(lokfile,orgfile);
+ lockfile=(char*)malloc(strlen(orgfile)+2);
+ strcpy(lockfile,orgfile);
 
- *lokfile='L';
+ *lockfile='L';
  *orgfile='D';
 
- if(!stat(lokfile,&buf))
+ if(!stat(lockfile,&buf))
     PrintMessage(Inform,"Lock file already exists for '%s'.",Url->name);
  else
    {
-    if(link(orgfile,lokfile))
-       PrintMessage(Warning,"Cannot make a lock file for '%s/%s/%s' [%!s].",Url->proto,Url->host,orgfile);
+    if(link(orgfile,lockfile))
+       PrintMessage(Warning,"Cannot make a lock file for '%s/%s/%s' [%!s].",Url->proto,Url->dir,orgfile);
    }
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
- free(lokfile);
+ free(lockfile);
  free(orgfile);
 }
 
@@ -792,28 +843,33 @@ int ExistsLockWebpageSpoolFile(URL *Url)
 {
  struct stat buf;
  char *file;
- int exists=0;
+ int existsL;
+
+ /* Check for configuration file option. */
+
+ if(!ConfigBoolean(LockFiles))
+    return(0);
 
  if(chdir(Url->proto))
     return(0);
 
- if(chdir(Url->host))
-   {chdir("..");return(0);}
+ if(chdir(Url->dir))
+   {fchdir(fSpoolDir);return(0);}
 
  /* Stat the file for the web page. */
 
  file=URLToFileName(Url);
- *file='L';
 
- exists=!stat(file,&buf);
+ *file='L';
+ existsL=!stat(file,&buf);
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
  free(file);
 
- return(exists);
+ return(existsL);
 }
 
 
@@ -825,30 +881,35 @@ int ExistsLockWebpageSpoolFile(URL *Url)
 
 void DeleteLockWebpageSpoolFile(URL *Url)
 {
- char *lokfile;
+ char *lockfile;
+
+ /* Check for configuration file option. */
+
+ if(!ConfigBoolean(LockFiles))
+    return;
 
  /* Change to the spool directory */
 
  if(chdir(Url->proto))
    {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",Url->proto);return;}
 
- if(chdir(Url->host))
-   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->host);chdir("..");return;}
+ if(chdir(Url->dir))
+   {PrintMessage(Warning,"Cannot change to directory '%s/%s' [%!s].",Url->proto,Url->dir);fchdir(fSpoolDir);return;}
 
  /* Delete the file for the backup web page. */
 
- lokfile=URLToFileName(Url);
+ lockfile=URLToFileName(Url);
 
- *lokfile='L';
+ *lockfile='L';
 
- if(unlink(lokfile))
-    PrintMessage(Inform,"Cannot unlink lock file '%s/%s/%s' [%!s].",Url->proto,Url->host,lokfile);
+ if(unlink(lockfile))
+    PrintMessage(Inform,"Cannot unlink lock file '%s/%s/%s' [%!s].",Url->proto,Url->dir,lockfile);
 
  /* Change dir back and tidy up. */
 
- chdir("../..");
+ fchdir(fSpoolDir);
 
- free(lokfile);
+ free(lockfile);
 }
 
 
@@ -866,7 +927,7 @@ int CreateLastTimeSpoolFile(URL *Url)
  char *file;
  int exists=0;
 
- if(NoLasttimeIndex)
+ if(ConfigBoolean(NoLasttimeIndex))
     return(1);
 
  /* Change to the last time directory */
@@ -879,17 +940,17 @@ int CreateLastTimeSpoolFile(URL *Url)
  file=URLToFileName(Url);
  *file='D';
 
- if(stat(file,&buf)==-1)
+ if(stat(file,&buf))
    {
-    char *name=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(file)+8);
-    sprintf(name,"../%s/%s/%s",Url->proto,Url->host,file);
+    char *name=(char*)malloc(strlen(Url->proto)+strlen(Url->dir)+strlen(file)+8);
+    sprintf(name,"../%s/%s/%s",Url->proto,Url->dir,file);
 
     if(link(name,file))
       {PrintMessage(Warning,"Cannot create file 'lasttime/%s' [%!s].",file);}
     else
       {
        *file='U';
-       sprintf(name,"../%s/%s/%s",Url->proto,Url->host,file);
+       sprintf(name,"../%s/%s/%s",Url->proto,Url->dir,file);
 
        if(link(name,file))
          {PrintMessage(Warning,"Cannot create file 'lasttime/%s' [%!s].",file);}
@@ -902,7 +963,7 @@ int CreateLastTimeSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(file);
 
@@ -910,41 +971,6 @@ int CreateLastTimeSpoolFile(URL *Url)
 }
 
 
-/*++++++++++++++++++++++++++++++++++++++
-  Determine if a particular URL exists in the last time spool directory.
-
-  int ExistsLastTimeSpoolFile Returns 1 if the URL exists.
-
-  URL *Url The URL to look for.
-  ++++++++++++++++++++++++++++++++++++++*/
-
-int ExistsLastTimeSpoolFile(URL *Url)
-{
- struct stat buf;
- int i,retval=0;
- char *name=URLToFileName(Url);
-
- *name='D';
-
- for(i=0;i<=NUM_PREVTIME_DIR;i++)
-   {
-    char lasttime[40];
-
-    if(i)
-       sprintf(lasttime,"prevtime%d/%s",i,name);
-    else
-       sprintf(lasttime,"lasttime/%s",name);
-
-    if(!stat(lasttime,&buf))
-      {retval=1;break;}
-   }
-
- free(name);
-
- return(retval);
-}
-
-
 /*++++++++++++++++++++++++++++++++++++++
   Delete a specified URL from the lasttime directory.
 
@@ -986,7 +1012,7 @@ char *DeleteLastTimeSpoolFile(URL *Url)
           unlink(name);
          }
 
-       chdir("..");
+       fchdir(fSpoolDir);
       }
    }
 
@@ -1006,12 +1032,10 @@ void CycleLastTimeSpoolFile(void)
  struct stat buf;
  int i;
 
- /* Cycle the directories */
+ /* Cycle the lasttime/prevtime? directories */
 
  for(i=NUM_PREVTIME_DIR;i>=0;i--)
    {
-    strcpy(prevlasttime,lasttime);
-
     if(i)
        sprintf(lasttime,"prevtime%d",i);
     else
@@ -1019,10 +1043,10 @@ void CycleLastTimeSpoolFile(void)
 
     /* Create it if it does not exist. */
 
-    if(lstat(lasttime,&buf))
+    if(stat(lasttime,&buf))
       {
        PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",lasttime);
-       if(mkdir(lasttime,0755))
+       if(mkdir(lasttime,(mode_t)ConfigInteger(DirPerm)))
           PrintMessage(Warning,"Cannot create directory '%s' [%!s].",lasttime);
       }
     else
@@ -1042,18 +1066,20 @@ void CycleLastTimeSpoolFile(void)
        dir=opendir(".");
 
        if(!dir)
-         {PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",lasttime);chdir("..");continue;}
+         {PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",lasttime);fchdir(fSpoolDir);continue;}
 
        ent=readdir(dir);  /* skip .  */
        if(!ent)
-         {PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",lasttime);closedir(dir);chdir("..");continue;}
+         {PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",lasttime);closedir(dir);fchdir(fSpoolDir);continue;}
        ent=readdir(dir);  /* skip .. */
 
        while((ent=readdir(dir)))
           if(unlink(ent->d_name))
              PrintMessage(Warning,"Cannot unlink previous time page '%s/%s' [%!s].",lasttime,ent->d_name);
 
-       chdir("..");
+       closedir(dir);
+
+       fchdir(fSpoolDir);
 
        if(rmdir(lasttime))
           PrintMessage(Warning,"Cannot unlink previous time directory '%s' [%!s].",lasttime);
@@ -1061,13 +1087,129 @@ void CycleLastTimeSpoolFile(void)
     else
        if(rename(lasttime,prevlasttime))
           PrintMessage(Warning,"Cannot rename previous time directory '%s' to '%s' [%!s].",lasttime,prevlasttime);
+
+    strcpy(prevlasttime,lasttime);
    }
 
- if(mkdir("lasttime",0755))
+ if(mkdir("lasttime",(mode_t)ConfigInteger(DirPerm)))
     PrintMessage(Warning,"Cannot create directory 'lasttime' [%!s].");
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Cycle the URLs from the outgoing directory to the lastout to prevout directories.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void CycleLastOutSpoolFile(void)
+{
+ char lastout[16],prevlastout[16];
+ struct stat buf;
+ int i;
+
+ /* Cycle the outgoing/lastout/prevout? directories */
+
+ for(i=NUM_PREVTIME_DIR;i>=0;i--)
+   {
+    if(i)
+       sprintf(lastout,"prevout%d",i);
+    else
+       strcpy(lastout,"lastout");
+
+    /* Create it if it does not exist. */
+
+    if(stat(lastout,&buf))
+      {
+       PrintMessage(Inform,"Directory '%s' does not exist [%!s]; creating one.",lastout);
+       if(mkdir(lastout,(mode_t)ConfigInteger(DirPerm)))
+          PrintMessage(Warning,"Cannot create directory '%s' [%!s].",lastout);
+      }
+    else
+       if(!S_ISDIR(buf.st_mode))
+          PrintMessage(Warning,"The file '%s' is not a directory.",lastout);
+
+    /* Delete the contents of the oldest one, rename the newer ones. */
+
+    if(i==NUM_PREVTIME_DIR)
+      {
+       DIR *dir;
+       struct dirent* ent;
+
+       if(chdir(lastout))
+         {PrintMessage(Warning,"Cannot change to directory '%s' [%!s].",lastout);continue;}
+
+       dir=opendir(".");
+
+       if(!dir)
+         {PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",lastout);fchdir(fSpoolDir);continue;}
+
+       ent=readdir(dir);  /* skip .  */
+       if(!ent)
+         {PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",lastout);closedir(dir);fchdir(fSpoolDir);continue;}
+       ent=readdir(dir);  /* skip .. */
+
+       while((ent=readdir(dir)))
+          if(unlink(ent->d_name))
+             PrintMessage(Warning,"Cannot unlink previous time page '%s/%s' [%!s].",lastout,ent->d_name);
+
+       closedir(dir);
+
+       fchdir(fSpoolDir);
+
+       if(rmdir(lastout))
+          PrintMessage(Warning,"Cannot unlink previous time directory '%s' [%!s].",lastout);
+      }
+    else
+       if(rename(lastout,prevlastout))
+          PrintMessage(Warning,"Cannot rename previous time directory '%s' to '%s' [%!s].",lastout,prevlastout);
+
+    strcpy(prevlastout,lastout);
+   }
+
+ if(mkdir("lastout",(mode_t)ConfigInteger(DirPerm)))
+    PrintMessage(Warning,"Cannot create directory 'lastout' [%!s].");
+
+ /* Link the files from the outgoing directory to the lastout directory. */
+
+ if(chdir("outgoing"))
+    PrintMessage(Warning,"Cannot change to directory 'outgoing' [%!s].");
+ else
+   {
+    DIR *dir;
+    struct dirent* ent;
+
+    dir=opendir(".");
+
+    if(!dir)
+       PrintMessage(Warning,"Cannot open current directory '%s' [%!s].",lastout);
+    else
+      {
+       ent=readdir(dir);  /* skip .  */
+       if(!ent)
+          PrintMessage(Warning,"Cannot read current directory '%s' [%!s].",lastout);
+       else
+         {
+          ent=readdir(dir);  /* skip .. */
+
+          while((ent=readdir(dir)))
+            {
+             char newname[40];
+
+             strcpy(newname,"../lastout/");
+             strcat(newname,ent->d_name);
+
+             if(link(ent->d_name,newname))
+                PrintMessage(Warning,"Cannot create lastout page '%s' [%!s].",&newname[3]);
+            }
+         }
+
+       closedir(dir);
+      }
+
+    fchdir(fSpoolDir);
+   }
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Open a file in the monitor directory to write into.
 
@@ -1092,10 +1234,10 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
 
  /* Create the monitor directory if needed and change to it */
 
- if(lstat("monitor",&buf))
+ if(stat("monitor",&buf))
    {
     PrintMessage(Inform,"Directory 'monitor' does not exist [%!s]; creating one.");
-    if(mkdir("monitor",0755))
+    if(mkdir("monitor",(mode_t)ConfigInteger(DirPerm)))
       {PrintMessage(Warning,"Cannot create directory 'monitor' [%!s].");return(-1);}
    }
  else
@@ -1111,7 +1253,8 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
 
  *file='O';
 
- fd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0644);
+ fd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+ init_buffer(fd);
 
  if(fd==-1)
    {PrintMessage(Warning,"Cannot create file 'monitor/%s' [%!s].",file);}
@@ -1120,7 +1263,8 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
     int ufd,mfd;
 
     *file='U';
-    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0644);
+    ufd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+    init_buffer(ufd);
 
     if(ufd!=-1)
       {
@@ -1134,7 +1278,8 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
       }
 
     *file='M';
-    mfd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0644);
+    mfd=open(file,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+    init_buffer(mfd);
 
     if(mfd!=-1)
       {
@@ -1153,7 +1298,7 @@ int CreateMonitorSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8],cha
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  free(file);
 
@@ -1198,7 +1343,7 @@ long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8]
  /* Check for 'O*' file. */
 
  *file='O';
- if(lstat(file,&buf))
+ if(stat(file,&buf))
     atime=0,mtime=time(NULL);
  else
     atime=buf.st_atime,mtime=buf.st_mtime;
@@ -1206,15 +1351,18 @@ long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8]
  /* Check for 'M*' file for the old format. */
 
  *file='M';
- if(lstat(file,&buf) || buf.st_size<8)
+ if(stat(file,&buf) || buf.st_size<8)
    {
     int interval=7;
 
     mfd=open(file,O_RDONLY|O_BINARY);
-    if(mfd)
+    init_buffer(mfd);
+
+    if(mfd!=-1)
       {
        char str[8];
-       if(read(mfd,str,8)>0)
+
+       if(read_data(mfd,str,8)>0)
           interval=atoi(str);
        if(interval<0)
           interval=0;
@@ -1241,31 +1389,34 @@ long ReadMonitorTimesSpoolFile(URL *Url,char MofY[13],char DofM[32],char DofW[8]
     else
        strcpy(HofD,"111111111111111111111111");
 
-    chdir("..");
+    fchdir(fSpoolDir);
     return(atime);
    }
 
  /* Assume that the file is in the new format. */
 
  mfd=open(file,O_RDONLY|O_BINARY);
- if(mfd)
+ init_buffer(mfd);
+
+ if(mfd!=-1)
    {
     char line[80];
+    line[78]=0;
 
-    if(read(mfd,line,80)>0)
-       if(sscanf(line,"%12s %31s %7s %24s",MofY,DofM,DofW,HofD)!=4 ||
-          strlen(MofY)!=12 || strlen(DofM)!=31 || strlen(DofW)!=7 || strlen(HofD)!=24)
-         {
-          strcpy(MofY,"111111111111");
-          strcpy(DofM,"1111111111111111111111111111111");
-          strcpy(DofW,"1111111");
-          strcpy(HofD,"100000000000000000000000");
-         }
+    if(read_data(mfd,line,80)!=78 ||
+       sscanf(line,"%12s %31s %7s %24s",MofY,DofM,DofW,HofD)!=4 ||
+       strlen(MofY)!=12 || strlen(DofM)!=31 || strlen(DofW)!=7 || strlen(HofD)!=24)
+      {
+       strcpy(MofY,"111111111111");
+       strcpy(DofM,"1111111111111111111111111111111");
+       strcpy(DofW,"1111111");
+       strcpy(HofD,"100000000000000000000000");
+      }
 
     close(mfd);
    }
 
- chdir("..");
+ fchdir(fSpoolDir);
  return(atime);
 }
 
@@ -1312,11 +1463,11 @@ char *DeleteMonitorSpoolFile(URL *Url)
     DIR *dir=opendir(".");
 
     if(!dir)
-      {err=PrintMessage(Warning,"Cannot open current directory 'monitor' [%!s].");chdir("..");return(err);}
+      {err=PrintMessage(Warning,"Cannot open current directory 'monitor' [%!s].");fchdir(fSpoolDir);return(err);}
 
     ent=readdir(dir);  /* skip .  */
     if(!ent)
-      {err=PrintMessage(Warning,"Cannot read current directory 'monitor' [%!s].");closedir(dir);chdir("..");return(err);}
+      {err=PrintMessage(Warning,"Cannot read current directory 'monitor' [%!s].");closedir(dir);fchdir(fSpoolDir);return(err);}
     ent=readdir(dir);  /* skip .. */
 
     while((ent=readdir(dir)))
@@ -1330,12 +1481,68 @@ char *DeleteMonitorSpoolFile(URL *Url)
 
  /* Change dir back and tidy up. */
 
- chdir("..");
+ fchdir(fSpoolDir);
 
  return(err);
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Create a temporary spool file.
+
+  int CreateTempSpoolFile Returns the file descriptor.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+int CreateTempSpoolFile(void)
+{
+ char name[16];
+ struct stat buf;
+ int fd;
+
+ /* Create the outgoing directory if needed and change to it */
+
+ if(stat("temp",&buf))
+   {
+    PrintMessage(Inform,"Directory 'temp' does not exist [%!s]; creating one.");
+    if(mkdir("temp",(mode_t)ConfigInteger(DirPerm)))
+      {PrintMessage(Warning,"Cannot create directory 'temp' [%!s].");return(-1);}
+   }
+ else
+    if(!S_ISDIR(buf.st_mode))
+      {PrintMessage(Warning,"The file 'temp' is not a directory.");return(-1);}
+
+ /* Open the file */
+
+ sprintf(name,"temp/tmp.%ld",(long)getpid());
+
+ fd=open(name,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,(mode_t)ConfigInteger(FilePerm));
+ init_buffer(fd);
+
+ if(fd==-1)
+    PrintMessage(Warning,"Cannot create temporary file '%s' [%!s].",name);
+
+ return(fd);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Close and delete the temporary spool file.
+
+  int fd The file descriptor.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void CloseTempSpoolFile(int fd)
+{
+ char name[16];
+
+ sprintf(name,"temp/tmp.%ld",(long)getpid());
+ if(unlink(name)==-1)
+    PrintMessage(Warning,"Cannot unlink temporary file '%s' [%!s].",name);
+
+ close(fd);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Convert a filename to a URL.
 
@@ -1360,6 +1567,7 @@ char *FileNameToURL(char *file)
  *copy='U';
 
  ufd=open(copy,O_RDONLY|O_BINARY);
+ init_buffer(ufd);
 
  if(ufd==-1)
    {
@@ -1368,7 +1576,7 @@ char *FileNameToURL(char *file)
     return(NULL);
    }
 
- r=read(ufd,path,256);
+ r=read_data(ufd,path,256);
 
  if(r==256)
    {
@@ -1378,7 +1586,7 @@ char *FileNameToURL(char *file)
        r+=rr;
        path=(char*)realloc(path,r+256+1);
       }
-    while((rr=read(ufd,&path[r],256))>0);
+    while((rr=read_data(ufd,&path[r],256))>0);
    }
  else if(r==-1)
    {
@@ -1419,3 +1627,26 @@ char *URLToFileName(URL *Url)
 
  return(file);
 }
+
+
+#if defined(__CYGWIN__)
+
+/*++++++++++++++++++++++++++++++++++++++
+  A fchdir replacement for systems that do not have one.
+
+  int fchdir Returns the error status from changing directory.
+
+  int fd The file descriptor of the spool directory (not used).
+  ++++++++++++++++++++++++++++++++++++++*/
+
+int fchdir(int fd)
+{
+ static char *spool_dir=NULL;
+
+ if(!spool_dir)
+    spool_dir=ConfigString(SpoolDir);
+
+ return(chdir(spool_dir));
+}
+
+#endif
